import platform
import numpy as np
import scipy as sp
import scipy.sparse
import ctypes
from ctypes import POINTER, c_int, c_char, c_char_p,  c_double, byref

try:
    if platform.system() == 'Windows':
        mkl = ctypes.windll.LoadLibrary("mkl_rt.dll")
    else:
        mkl = ctypes.cdll.LoadLibrary("libmkl_rt.dylib")
except:
    raise ImportError("Could not load Intel MKL Library.")


def mkl_csr_matvec(A, x, transpose=False):
    """
    Parameters
    ----------
    A : scipy.sparse csr matrix
    x : numpy 1d array
    """

    if not sp.sparse.isspmatrix_csr(A):
        raise TypeError("The matrix must be a scipy sparse CSR matrix.")

    if x.ndim != 1:
        raise TypeError("The vector to be multiplied must be a 1d array.")

    if x.dtype.type is not np.double:
        x = x.astype(np.double, copy=True)

    # Allocate the result of the matrix-vector multiplication.
    result = np.empty(A.shape[transpose])

    # Set the parameters for simply computing A.dot(x) for a general matrix A.
    alpha = byref(c_double(1.0))
    beta = byref(c_double(0.0))
    matrix_description = c_char_p(bytes('G  C  ', 'utf-8'))

    # Get pointers to the numpy arrays.
    data_ptr = A.data.ctypes.data_as(POINTER(c_double))
    indices_ptr = A.indices.ctypes.data_as(POINTER(c_int))
    indptr_begin = A.indptr[:-1].ctypes.data_as(POINTER(c_int))
    indptr_end = A.indptr[1:].ctypes.data_as(POINTER(c_int))
    x_ptr = x.ctypes.data_as(POINTER(c_double))
    result_ptr = result.ctypes.data_as(POINTER(c_double))

    transpose_flag = byref(c_char(bytes(['n', 't'][transpose], 'utf-8')))
    n_row, n_col = [byref(c_int(size)) for size in A.shape]
    mkl.mkl_dcsrmv(
        transpose_flag, n_row, n_col, alpha, matrix_description,
        data_ptr, indices_ptr, indptr_begin, indptr_end, x_ptr, beta, result_ptr
    )
    return result



from warnings import warn
import numpy as np
import scipy.sparse as sparse
from .abstract_matrix import AbstractDesignMatrix
try:
    from .mkl_matvec import mkl_csr_matvec
except:
    mkl_csr_matvec = None


class SparseDesignMatrix(AbstractDesignMatrix):

    def __init__(self, X, use_mkl=True, center_predictor=False, add_intercept=True,
                 copy_array=False, dot_format='csr', Tdot_format='csr'):
        """
        Params:
        ------
        X : scipy sparse matrix
        """
        if copy_array:
            X = X.copy()
        super().__init__()
        if dot_format == 'csc' or Tdot_format == 'csc':
            raise NotImplementedError(
                "Current dot operations are only implemented for the CSR format."
            )
        X = X.tocsr()
        X = self.remove_intercept_indicator(X)

        if use_mkl and (mkl_csr_matvec is None):
            warn("Could not load MKL Library. Will use Scipy's 'dot'.")
            use_mkl = False
        self.use_mkl = use_mkl

        self.centered = center_predictor
        if center_predictor:
            self.column_offset = np.squeeze(np.array(X.mean(axis=0)))
        else:
            self.column_offset = np.zeros(X.shape[1])

        self.intercept_added = add_intercept
        self.X_main = X.tocsr()

    @property
    def shape(self):
        shape = self.X_main.shape
        return shape[0], shape[1] + int(self.intercept_added)

    @property
    def is_sparse(self):
        return True

    @property
    def nnz(self):
        """ Currently exists only to estimate the cost of matrix-matrix and
        matrix-vector operations. Does not correspond to the actual nnz of the
        represented design matrix.
        """
        return self.X_main.nnz

    def dot(self, v):

        if self.memoized:
            if np.all(self.v_prev == v):
                return self.X_dot_v
            self.v_prev = v.copy()

        intercept_effect = 0.
        if self.intercept_added:
            intercept_effect += v[0]
            v = v[1:]
        result = intercept_effect + self.main_dot(v)
        
        if self.memoized:
            self.X_dot_v = result
        self.dot_count += 1

        return result

    def main_dot(self, v):
        """ Multiply by the main effect part of the design matrix. """
        X = self.X_main
        result = mkl_csr_matvec(X, v) if self.use_mkl else X.dot(v)
        result -= np.inner(self.column_offset, v)
        if self.memoized:
            self.X_dot_v = result
        return result

    def Tdot(self, v):
        result = self.main_Tdot(v)
        if self.intercept_added:
            result = np.concatenate(([np.sum(v)], result))
        self.Tdot_count += 1
        return result

    def main_Tdot(self, v):
        X = self.X_main
        result = mkl_csr_matvec(X, v, transpose=True) \
            if self.use_mkl else X.T.dot(v)
        result -= np.sum(v) * self.column_offset
        return result

    def compute_fisher_info(self, weight, diag_only=False):
        """ Compute $X^T W X$ where W is the diagonal matrix of a given weight."""

        if diag_only:
            return self.compute_fisher_diag(weight)

        weight_mat = self.create_diag_matrix(weight)
        X = self.X_main
        X_T = X.T
        weighted_X = weight_mat.dot(X).tocsc()

        n_pred = self.shape[1]
        fisher_info = np.zeros((n_pred, n_pred))
        if self.intercept_added:
            fisher_info[0, 0] = np.sum(weight)
            fisher_info[0, 1:] \
                = weighted_X.sum(0) - np.sum(weight) * self.column_offset
            fisher_info[1:, 0] = fisher_info[0, 1:]
            fisher_info_wo_intercept = fisher_info[1:, 1:]
        else:
            fisher_info_wo_intercept = fisher_info

        fisher_info_wo_intercept += X_T.dot(weighted_X).toarray()
        if self.centered:
            outer_prod_term = np.outer(
                self.column_offset, weighted_X.sum(0)
            )
            fisher_info_wo_intercept -= outer_prod_term + outer_prod_term.T
            fisher_info_wo_intercept \
                += np.sum(weight) * np.outer(self.column_offset, self.column_offset)

        return fisher_info

    def compute_fisher_diag(self, weight):

        weight_mat = self.create_diag_matrix(weight)
        diag = weight_mat.dot(self.X_main.power(2)).sum(0)
        if self.centered:
            weighted_X = weight_mat.dot(self.X_main).tocsc()
            diag -= 2 * self.column_offset \
                    * np.squeeze(np.asarray(weighted_X.sum(0)))
            diag += np.sum(weight) * self.column_offset ** 2
        diag = np.squeeze(np.asarray(diag))
        if self.intercept_added:
            diag = np.concatenate(([np.sum(weight)], diag))

        return diag

    def create_diag_matrix(self, v):
        return sparse.dia_matrix((v, 0), (len(v), len(v)))

    def toarray(self):
        X = self.X_main.toarray() - self.column_offset[np.newaxis, :]
        if self.intercept_added:
            X = np.hstack((np.ones(X.shape[0]), X))
        return X

    def extract_matrix(self, order=None):
        pass



from warnings import warn
import scipy as sp

from .linear_model import LinearModel
from .logistic_model import LogisticModel
from .cox_model import CoxModel
from ..design_matrix import DenseDesignMatrix, SparseDesignMatrix

def RegressionModel(
        outcome, X, family='linear',
        add_intercept=None, center_predictor=True
    ):
    """ Prepare input data to BayesBridge, with pre-processings as needed.

    For the Cox model, the observations (rows of X) are reordered to optimize
    likelihood, gradient, and Hessian evaluations.

    Parameters
    ----------
    outcome : 1-d numpy array, tuple of two 1-d numpy arrays
        n_success or (n_success, n_trial) if family == 'logistic'. If
        the input is a single array, then outcome is assumed binary.
        (event_time, censoring_time) if family == 'cox'.
    X : numpy array or scipy sparse matrix
    family : str, {'linear', 'logit', 'cox'}
    add_intercept : bool, None
        If None, add intercept except when family == 'cox'
    center_predictor : bool
    """

    if add_intercept is None:
        add_intercept = (family != 'cox')

    if family == 'cox':
        if add_intercept:
            add_intercept = False
            warn("Intercept is not identifiable in Cox model and won't be added.")
        event_time, censoring_time = outcome
        event_time, censoring_time, X = CoxModel.preprocess_data(
            event_time, censoring_time, X
        )

    is_sparse = sp.sparse.issparse(X)
    DesignMatrix = SparseDesignMatrix if is_sparse else DenseDesignMatrix
    design = DesignMatrix(
        X, add_intercept=add_intercept, center_predictor=center_predictor
    )

    if family == 'linear':
        model = LinearModel(outcome, design)
    elif family == 'logit':
        if isinstance(outcome, tuple):
            n_success, n_trial = outcome
        else:
            n_success = outcome
            n_trial = None
        model = LogisticModel(n_success, n_trial, design)
    elif family == 'cox':
        model = CoxModel(event_time, censoring_time, design)
    else:
        raise NotImplementedError()

    return model



# cython: cdivision = True
from libc.math cimport exp, log, sqrt, fabs, M_PI
import random
import cython
import numpy as np
cimport numpy as np
from numpy.random import PCG64
from numpy.random.bit_generator cimport BitGenerator
from scipy_ndtr cimport log_ndtr as normal_logcdf
from bayesbridge.random.normal.normal cimport random_normal
from bayesbridge.random.uniform.uniform cimport random_uniform



cdef class PolyaGammaDist():
    # Threshold below (and above) which the target density is bounded by inverse
    # Gaussian (and exponential) and have different analytical series expressions.
    cdef double THRESHOLD
    # Number of terms in the infinite alternating series beyond which to truncate.
    cdef int MAX_SERIES_TERMS
    cdef BitGenerator bitgen

    def __init__(self, seed=None):
        self.set_seed(seed)
        self.THRESHOLD = 2.0 / M_PI
        self.MAX_SERIES_TERMS = 100
        self.bitgen = PCG64(seed)

    def set_seed(self, seed):
        self.bitgen = PCG64(seed)

    def get_state(self):
        return self.bitgen.state

    def set_state(self, state):
        self.bitgen.state = state

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def rand_polyagamma(self, shape, tilt):
        """
        Sample from exponentially tilted Polya-Gamma distribution
            p(x | shape, tilt) \propto \exp(- tilt^2 / 2 * x) p(x | shape, tilt=0)
        via Devroye's alternatig series method.

        Parameters
        ----------
        shape : numpy array of integers
            Distribution is defined for non-integer values but the implemented
            algorithm only support integer values.
        tilt : numpy array of doubles
            Sqrt of twice the negative tilting parameter, actually.
        """

        if not isinstance(shape, np.ndarray) and isinstance(tilt, np.ndarray):
            raise TypeError('Input must be numpy arrays.')
        if not shape.size == tilt.size:
            raise ValueError('Input arrays must be of the same length.')
        if not np.issubdtype(shape.dtype, np.integer):
            raise ValueError('Shape parameter must be integers.')
        shape = shape.astype(np.intc)
        tilt = tilt.astype(np.double)
        result = np.zeros(shape.size, dtype=np.double)

        cdef int[:] shape_view = shape
        cdef double[:] tilt_view = tilt
        cdef double[:] result_view = result
        cdef long n_samples = shape_view.size
        cdef Py_ssize_t index, j
        for index in range(n_samples):
            for j in range(shape_view[index]):
                result_view[index] \
                    += self.rand_scalar_unit_shape_polyagamma(tilt_view[index])
        return result

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def rand_unit_shape_polyagamma(self, tilt):

        if not isinstance(tilt, np.ndarray):
            raise TypeError('Input must be numpy arrays.')
        tilt = tilt.astype(np.double)
        result = np.empty(tilt.size, dtype=np.double)

        cdef double[:] tilt_view = tilt
        cdef double[:] result_view = result
        cdef long n_samples = tilt_view.size
        cdef Py_ssize_t index
        for index in range(n_samples):
            result_view[index] \
                = self.rand_scalar_unit_shape_polyagamma(tilt_view[index])
        return result

    cdef double rand_scalar_unit_shape_polyagamma(self, double tilt):
        return .25 * self.rand_tilted_jocobi(.5 * fabs(tilt))

    cdef double rand_tilted_jocobi(self, double tilt):
        """
        Sample from tilted Jacobi distribution
            p(x | tilt) \propto \exp(- tilt^2 / 2 * x) p(x | 0)
        via Devroye's alternatig series method.
        """
        cdef double X, U, proposal_density
        cdef bint accepted = False

        # Main sampling loop page 130 of the Windle PhD thesis
        while not accepted:
            X, proposal_density = self.rand_proposal(tilt)
            U = random_uniform(self.bitgen) * proposal_density
            accepted = self.decide_acceptability(U, X, proposal_density)

        return X

    cdef (double, double) rand_proposal(self, double tilt):
        # Many quantities here can be cached and reused in case of rejection, but
        # the acceptance rate is so high that it does not matter.
        cdef double exp_rate = .5 * tilt ** 2 + .125 * M_PI ** 2
        cdef double prob_to_right = self.calc_prob_to_right(tilt, exp_rate)
        if random_uniform(self.bitgen) < prob_to_right:
            X = self.rand_left_truncated_exp(1. / exp_rate, self.THRESHOLD)
        else:
            X = self.rand_right_truncated_unit_shape_invgauss(tilt, self.THRESHOLD)
        proposal_density = self.calc_next_term_in_series(0, X)
        return X, proposal_density

    cdef double calc_prob_to_right(self, double tilt, double exp_rate):
        cdef double log_mass_expo \
            = - log(exp_rate) - exp_rate * self.THRESHOLD + log(.25 * M_PI)
        cdef double log_mass_invg_1 \
            = - tilt + normal_logcdf(
                (self.THRESHOLD * tilt - 1.) / sqrt(self.THRESHOLD))
        cdef double log_mass_invg_2 \
            = tilt + normal_logcdf(
                - (self.THRESHOLD * tilt + 1.) / sqrt(self.THRESHOLD))
        cdef double mass_ratio = (
            exp(log_mass_invg_1 - log_mass_expo)
            + exp(log_mass_invg_2 - log_mass_expo)
        )
        return 1.0 / (1.0 + mass_ratio)

    # Equations (12) and (13) of Polson, Scott, and Windle (2013)
    cdef double calc_next_term_in_series(self, int n, double x):
        cdef double log_result = log(M_PI * (n + 0.5))
        if x <= self.THRESHOLD:
            log_result += - 1.5 * log(.5 * x * M_PI) - 2 * (n + 0.5) ** 2 / x
        else:
            log_result += - 0.5 * x * M_PI ** 2 * (n + 0.5) ** 2
        return exp(log_result)

    cdef bint decide_acceptability(self, double U, double X, double zeroth_term):

        cdef double partial_sum = zeroth_term
        cdef int n_summed = 1
        cdef int sign = -1 # Sign of the next term in the alternating sequence
        cdef bint acceted
        cdef bint is_determinate = False

        while not is_determinate:
            partial_sum += sign * self.calc_next_term_in_series(n_summed, X)
            n_summed += 1
            if sign == -1:
                if U <= partial_sum:
                    accepted = True
                    is_determinate = True
            else: # sign == 1
                if U > partial_sum:
                    accepted = False
                    is_determinate = True
                elif n_summed >= self.MAX_SERIES_TERMS:
                    acceted = True # Take the partial sum lower-bound as the target
                    is_determinate = True
            sign = - sign

        return accepted

    cdef double rand_left_truncated_exp(self, double scale, double trunc):
        return trunc - scale * log(1.0 - random_uniform(self.bitgen))

    # Ref: "Simulation of truncated gamma variables" by Younshik Chung
    # Korean Journal of Computational & Applied Mathematics, 1998
    cdef double rand_left_truncated_chisq(self, double trunc):
        cdef double X, density_ratio
        cdef bint accepted = False
        while not accepted:
            X = self.rand_left_truncated_exp(2., trunc)
            density_ratio = sqrt(0.5 * M_PI / X)
            accepted = (random_uniform(self.bitgen) <= density_ratio)
        return X


    cdef double rand_right_truncated_unit_shape_invgauss(self, double rate, double trunc):
        # Shape parameter is assumed to be one.
        cdef double X
        cdef double mean = 1. / rate
        cdef bint accepted = False

        # Choose a better sampler depending on the input parameters
        if mean > trunc:
            # Algorithm 3 in Windle's PhD thesis, page 128
            while not accepted:
                X = 1.0 / self.rand_left_truncated_chisq(.5 * M_PI)
                accepted = (log(random_uniform(self.bitgen)) < - 0.5 * X * rate ** 2)
        else:
            while not accepted:
                X = self.rand_unit_shape_invgauss(mean)
                accepted = (X < trunc)
        return X

    cdef double rand_unit_shape_invgauss(self, double mean):
        cdef double V = random_normal(self.bitgen) ** 2
        cdef double X = mean + 0.5 * mean * (
            mean * V - sqrt(4.0 * mean * V + mean ** 2 * V ** 2)
        )
        if random_uniform(self.bitgen) > mean / (mean + X):
            X = mean ** 2 / X
        return X




from .abstract_model import AbstractModel
import numpy as np
import numpy.random
from warnings import warn

class LogisticModel(AbstractModel):

    # TODO: Python crushes during the Gibbs if n_success has the second
    # dimension (instead of being a vector). Add checks for the inputs.
    def __init__(self, n_success, n_trial, design):

        self.check_input_validity(n_success, n_trial, design)
        if n_trial is None:
            n_trial = np.ones(len(n_success))
            warn(
                "The numbers of trials were not specified. The binary "
                "outcome is assumed."
            )

        self.n_trial = n_trial.astype('float64')
        self.n_success = n_success.astype('float64')
        self.design = design
        self.name = 'logit'

    def check_input_validity(self, n_success, n_trial, design):

        if n_trial is None:
            if np.max(n_success) > 1:
                raise ValueError(
                    "If not binary, the number of trials must be specified.")
            if not len(n_success) == design.shape[0]:
                raise ValueError(
                    "Incompatible sizes of the outcome and design matrix."
                )
            return # No need to check the rest for the default initialization.

        if not len(n_trial) == len(n_success) == design.shape[0]:
            raise ValueError(
                "Incompatible sizes of the outcome vectors and design matrix."
            )

        if np.any(n_trial <= 0):
            raise ValueError("Number of trials must be strictly positive.")

        if np.any(n_success > n_trial):
            raise ValueError(
                "Number of successes cannot be larger than that of trials.")

    def compute_loglik_and_gradient(self, beta, loglik_only=False):
        logit_prob = self.design.dot(beta)
        predicted_prob = LogisticModel.convert_to_probability_scale(logit_prob)
        loglik = np.sum(
            self.n_success * logit_prob \
            - self.n_trial * np.log(1 + np.exp(logit_prob))
        )
        if loglik_only:
            grad = None
        else:
            grad = self.design.Tdot(self.n_success - self.n_trial * predicted_prob)
        return loglik, grad

    def compute_hessian(self, beta):
        predicted_prob = LogisticModel.compute_predicted_prob(self.design, beta)
        weight = predicted_prob * (1 - predicted_prob)
        return - self.design.compute_fisher_info(weight)

    def get_hessian_matvec_operator(self, beta):
        predicted_prob = LogisticModel.compute_predicted_prob(self.design, beta)
        weight = predicted_prob * (1 - predicted_prob)
        hessian_op = lambda v: \
            - self.design.Tdot(self.n_trial * weight * self.design.dot(v))
        return hessian_op

    def calc_intercept_mle(self):
        binom_prob_mle = self.n_success.mean() / self.n_trial.mean()
        intercept = np.log(binom_prob_mle / (1 - binom_prob_mle))
        return intercept

    @staticmethod
    def compute_polya_gamma_mean(shape, tilt):
        min_magnitude = 1e-5
        pg_mean = shape.copy() / 2
        is_nonzero = (np.abs(tilt) > min_magnitude)
        pg_mean[is_nonzero] \
            *= 1 / tilt[is_nonzero] \
               * (np.exp(tilt[is_nonzero]) - 1) / (np.exp(tilt[is_nonzero]) + 1)
        return pg_mean

    @staticmethod
    def compute_predicted_prob(X, beta, truncate=False):
        logit_prob = X.dot(beta)
        return LogisticModel.convert_to_probability_scale(logit_prob, truncate)

    @staticmethod
    def convert_to_probability_scale(logit_prob, truncate=False):
        # The flag 'truncate == True' guarantees 0 < prob < 1.
        if truncate:
            upper_bd = 36.7  # approximately - log(2 ** -53)
            lower_bd = - 709  # approximately - log(2 ** 1023)
            logit_prob[logit_prob > upper_bd] = upper_bd
            logit_prob[logit_prob < lower_bd] = lower_bd
        prob = 1 / (1 + np.exp(-logit_prob))
        return prob

    @staticmethod
    def simulate_outcome(n_trial, X, beta, seed=None):
        """
        Parameters
        ----------
        X : DesignMatrix, numpy/scipy matrix
            Only needs to support the `dot()` operation
        """
        prob = LogisticModel.compute_predicted_prob(X, beta)
        if seed is not None:
            np.random.seed(seed)
        y = np.random.binomial(n_trial, prob)
        return y


import math
import numpy as np
import scipy as sp
import scipy.sparse
from warnings import warn
from functools import partial
from .cg_sampler import ConjugateGradientSampler
from .reg_coef_posterior_summarizer import RegressionCoeffficientPosteriorSummarizer
from .direct_gaussian_sampler import generate_gaussian_with_weight
from .hamiltonian_monte_carlo import hmc
from .hamiltonian_monte_carlo.nuts import NoUTurnSampler
from .hamiltonian_monte_carlo.stepsize_adapter \
    import HamiltonianBasedStepsizeAdapter


class SparseRegressionCoefficientSampler():

    def __init__(self, n_coef, prior_sd_for_unshrunk, sampling_method,
                 stability_estimate_stabilized=False,
                 regularizing_slab_size=float('inf')):

        self.prior_sd_for_unshrunk = prior_sd_for_unshrunk
        self.n_unshrunk = len(prior_sd_for_unshrunk)
        self.regularizing_slab_size = regularizing_slab_size

        # Object for keeping track of running average.
        self.regcoef_summarizer = RegressionCoeffficientPosteriorSummarizer(
            n_coef, self.n_unshrunk, regularizing_slab_size,
            pc_summary_method='average'
        )
        if sampling_method == 'cg':
            self.cg_sampler = ConjugateGradientSampler(self.n_unshrunk)
        elif sampling_method in ['hmc', 'nuts']:
            self.stability_adjustment_adapter = \
                HamiltonianBasedStepsizeAdapter(init_stepsize=.3, target_accept_prob=.95)
            self.stability_est_stabilizer = StabilityEstimateStabilizer()
            self.stability_est_stabilized = stability_estimate_stabilized
        self._sampling_info_attributes = [
            'regcoef_summarizer',
            'stability_adjustment_adapter',
            'stability_estimate_stabilizer'
        ] # Names of the attributes tracking info from previous sampling iterations.

    def get_internal_state(self):
        state = {}
        for attr in self._sampling_info_attributes:
            if hasattr(self, attr):
                state[attr] = getattr(self, attr)
        return state

    def set_internal_state(self, state):
        for attr in self._sampling_info_attributes:
            if hasattr(self, attr):
                setattr(self, attr, state[attr])

    def sample_gaussian_posterior(
            self, y, design, obs_prec, gscale, lscale, method='cg'):
        """
        Parameters
        ----------
        beta_init: vector
            Used when when method == 'cg' as the starting value of the
            preconditioned conjugate gradient algorithm.
        method: {'cholesky', 'cg'}
            If 'cholesky', a sample is generated using a cholesky method based on the
            cholesky linear algebra. If 'cg', the preconditioned conjugate gradient
            sampler is used.
        """
        # TODO: Comment on the form of the posterior.

        v = design.Tdot(obs_prec * y)
        prior_shrunk_scale = self.compute_prior_shrunk_scale(gscale, lscale)
        prior_sd = np.concatenate((
            self.prior_sd_for_unshrunk, prior_shrunk_scale
        ))
        prior_prec_sqrt = 1 / prior_sd

        info = {}
        if method == 'cholesky':
            beta = generate_gaussian_with_weight(
                design, obs_prec, prior_prec_sqrt, v)

        elif method == 'cg':
            beta_condmean_guess = \
                self.regcoef_summarizer.extrapolate_beta_condmean(gscale, lscale)
            beta_precond_scale_sd = self.regcoef_summarizer.estimate_beta_precond_scale_sd()
            beta, cg_info = self.cg_sampler.sample(
                design, obs_prec, prior_prec_sqrt, v,
                beta_init=beta_condmean_guess,
                precond_by='prior',
                beta_scaled_sd=beta_precond_scale_sd,
                maxiter=500, atol=10e-6 * np.sqrt(design.shape[1])
            )
            self.regcoef_summarizer.update(beta, gscale, lscale)
            info['n_cg_iter'] = cg_info['n_iter']

        else:
            raise NotImplementedError()

        return beta, info

    def sample_by_hmc(
            self, beta, gscale, lscale, model, method='hmc', max_step=512):
        # TODO: allow for a fixed stepsize (w/o adaptation)?

        beta_precond_post_sd = \
            self.regcoef_summarizer.estimate_beta_precond_scale_sd()
        precond_scale, precond_prior_prec = \
            self.compute_preconditioning_scale(
                gscale, lscale, beta_precond_post_sd, self.prior_sd_for_unshrunk
            )

        # Calibrate the stepsize by estimating the stability limit.
        approx_stability_limit, n_hessian_matvec \
            = self.compute_stability_limit(
                gscale, lscale, model, precond_scale, precond_prior_prec
            )
        if self.stability_est_stabilized:
            pre_stabilization = approx_stability_limit
            approx_stability_limit = \
                self.stability_est_stabilizer.stabilize(approx_stability_limit)
            self.stability_est_stabilizer.update(pre_stabilization)
        adjustment_factor = self.stability_adjustment_adapter.get_current_stepsize()
        stepsize_upper_limit = adjustment_factor * approx_stability_limit
        dt = np.random.uniform(.5, 1) * stepsize_upper_limit

        # Sample the coefficients.
        beta_precond = beta / precond_scale
        f = self.get_precond_logprob_and_gradient(
            model, precond_scale, precond_prior_prec
        )
        if method == 'hmc':
            integration_time = np.pi / 2 * np.random.uniform(.8, 1.)
            n_step = np.ceil(integration_time / dt).astype('int')
            n_step = min(n_step, max_step)
            beta_precond, hmc_info = \
                hmc.generate_next_state(f, dt, n_step, beta_precond)
            info = {
                key: hmc_info.get(key)
                for key in ['accepted', 'accept_prob', 'n_grad_evals',
                            'instability_detected']
            }
            info['n_integrator_step'] = n_step
            hamiltonian_error = hmc_info['hamiltonian_error']

        elif method == 'nuts':
            nuts = NoUTurnSampler(f)
            max_height = int(math.log2(max_step))
            beta_precond, nuts_info = \
                nuts.generate_next_state(dt, beta_precond, max_height=max_height)
            info = {
                key: nuts_info.get(key)
                for key in ['ave_accept_prob', 'n_grad_evals', 'tree_height',
                            'instability_detected']
            }
            hamiltonian_error = nuts_info['ave_hamiltonian_error']

        else:
            raise NotImplementedError()

        beta = beta_precond * precond_scale
        self.regcoef_summarizer.update(beta, gscale, lscale)
        self.stability_adjustment_adapter.adapt_stepsize(hamiltonian_error)

        info['n_hessian_matvec'] = n_hessian_matvec
        info['stepsize'] = dt
        info['stability_limit_est'] = approx_stability_limit
        info['stability_adjustment_factor'] = adjustment_factor
        return beta, info

    def compute_preconditioning_scale(
            self, gscale, lscale, regcoef_precond_post_sd,
            prior_sd_for_unshrunk, unshrunk_target_sd_scale=1.):
        n_coef = len(regcoef_precond_post_sd)
        n_unshrunk = n_coef - len(lscale)

        precond_scale = np.ones(n_coef)
        precond_scale[n_unshrunk:] \
            = self.compute_prior_shrunk_scale(gscale, lscale)
        if n_unshrunk > 0:
            precond_scale[:n_unshrunk] = \
                unshrunk_target_sd_scale * regcoef_precond_post_sd[:n_unshrunk]

        precond_prior_prec = np.concatenate((
            (prior_sd_for_unshrunk / precond_scale[:n_unshrunk]) ** -2,
            np.ones(len(lscale))
        ))

        return precond_scale, precond_prior_prec

    def compute_prior_shrunk_scale(self, gscale, lscale):
        """ Compute the prior scale for the coefficient under regularized
        shrinkage in a numerically stable way. """
        prior_shrunk_scale = gscale * lscale
        prior_shrunk_scale /= np.sqrt(
            1 + (prior_shrunk_scale / self.regularizing_slab_size) ** 2
        )
        return prior_shrunk_scale

    def compute_stability_limit(
            self, gscale, lscale, model, precond_scale, precond_prior_prec):
        beta_condmean_guess = \
            self.regcoef_summarizer.extrapolate_beta_condmean(gscale, lscale)
        hessian_pc_estimate = self.regcoef_summarizer.estimate_precond_hessian_pc()
        max_curvature, hessian_pc, n_hessian_matvec = \
            self.compute_precond_hessian_curvature(
                beta_condmean_guess, model, precond_scale, precond_prior_prec,
                hessian_pc_estimate
            )
        self.regcoef_summarizer.update_precond_hessian_pc(hessian_pc)
        approx_stability_limit = 2 / np.sqrt(max_curvature)
        return approx_stability_limit, n_hessian_matvec

    def compute_precond_hessian_curvature(
            self, beta_location, model, precond_scale, precond_prior_prec, pc_estimate):

        precond_hessian_matvec, matvec_counter = self.get_precond_hessian_matvec(
            model, beta_location, precond_scale, precond_prior_prec
        )
        precond_hessian_op = sp.sparse.linalg.LinearOperator(
            (len(beta_location), len(beta_location)), precond_hessian_matvec
        )
        if pc_estimate is None:
            pc_estimate = np.random.randn(len(beta_location))
        eigval, eigvec = sp.sparse.linalg.eigsh(
            precond_hessian_op, k=1, tol=.1, v0=pc_estimate, ncv=2
        )   # We don't need a high (relative) accuracy.
        max_curvature = eigval[0]
        pc = np.squeeze(eigvec)
        if max_curvature <= 0:
            raise ArithmeticError(
                "Numerical instability occured during the Lancoz iteration and "
                "a negative curvature value returned for a log-concave distribution. "
                "Likely caused by divergence of regression coefficients to infinity. "
                "Check the input data and / or place an informative prior."
            )
        return max_curvature, pc, matvec_counter[0]

    @staticmethod
    def get_precond_hessian_matvec(
            model, beta_location, precond_scale, precond_prior_prec, obs_prec=None):

        if model.name == 'linear':
            args = [beta_location, obs_prec]
        else:
            args = [beta_location]
        loglik_hessian_matvec = model.get_hessian_matvec_operator(*args)
        matvec_counter = [0]
        def precond_hessian_matvec(beta_precond):
            matvec_counter[0] += 1
            return precond_prior_prec * beta_precond \
                   - precond_scale * loglik_hessian_matvec(precond_scale * beta_precond)

        return precond_hessian_matvec, matvec_counter

    @staticmethod
    def get_precond_logprob_and_gradient(
            model, precond_scale, precond_prior_prec, obs_prec=None):

        compute_loglik_and_gradient = model.compute_loglik_and_gradient
        if model.name == 'linear':
            compute_loglik_and_gradient \
                = partial(compute_loglik_and_gradient, obs_prec=obs_prec)
        def f(beta_precond, loglik_only=False):
            beta = beta_precond * precond_scale
            logp, grad_wrt_beta = \
                compute_loglik_and_gradient(beta, loglik_only=loglik_only)
            logp += np.sum(- precond_prior_prec * beta_precond ** 2) / 2

            grad = None
            if not loglik_only and math.isfinite(logp):
                grad = precond_scale * grad_wrt_beta  # Chain rule.
                grad += - precond_prior_prec * beta_precond
            return logp, grad

        return f

    def search_mode(self, beta, lscale, gscale, obs_prec, model, optim_maxiter=None,
                    use_newton_method=False, require_trust_region=False,
                    warn_optim_failure=False):

        if (not use_newton_method) and require_trust_region:
            warn("Trust regions are used only for Newton methods.")

        precond_scale, compute_negative_logp, compute_negative_grad, precond_hessian_matvec \
            = self.define_function_for_optim(beta, lscale, gscale, obs_prec, model)
        if not use_newton_method:
            precond_hessian_matvec = None

        optim_method, optim_options = self.choose_optim_method_and_options(
            optim_maxiter, use_newton_method, require_trust_region, n_param=len(beta)
        )

        beta_precond = beta / precond_scale
        model.design.memoize_dot(True)
        model.design.reset_matvec_count()
            # Avoid matrix-vector multiplication with the same input.
        optim_result = sp.optimize.minimize(
            compute_negative_logp, beta_precond, method=optim_method,
            jac=compute_negative_grad, hessp=precond_hessian_matvec,
            options=optim_options
        )
        model.design.memoize_dot(False)
        if (not optim_result.success) and warn_optim_failure:
            warn(
                "The regression coefficient mode (conditionally on the scale "
                "parameters could not be located within {:d} iterations of optimization "
                "steps. Proceeding with the current best estimate.".format(
                    optim_result.nit
                )
            )
        beta = precond_scale * optim_result.x
        info = {
            'is_success': optim_result.success,
            'method': optim_method,
            'n_iter': optim_result['nit'],
            'n_logp_eval': optim_result['nfev'],
            'n_grad_eval': optim_result.get('njev', 0),
                # incorrect output as of the current Scipy version (to be fixed in ver. 1.3.0)
            'n_hess_eval': optim_result.get('nhev', 0),
                # incorrect output as of the current Scipy version (to be fixed in ver. 1.3.0)
            'n_design_matvec': model.design.n_matvec,
        }
        return beta, info

    def define_function_for_optim(self, beta, lscale, gscale, obs_prec, model):

        beta_precond_post_sd = np.ones(beta.size)
        # No Monte Carlo estimate yet, so make some reasonable guess. It
        # probably should depend on the outcome and design matrix.
        precond_scale, precond_prior_prec = \
            self.compute_preconditioning_scale(
                gscale, lscale, beta_precond_post_sd, self.prior_sd_for_unshrunk
            )

        f = self.get_precond_logprob_and_gradient(
            model, precond_scale, precond_prior_prec, obs_prec
        )
        def compute_negative_logp(beta_precond):
            # Negative log-density
            return - f(beta_precond, loglik_only=True)[0]

        def compute_negative_grad(beta_precond):
            return - f(beta_precond)[1]


        def precond_hessian_matvec(precond_location, v):
            hessian_eval_location = precond_scale * precond_location
            hessian_matvec, _ = self.get_precond_hessian_matvec(
                model, hessian_eval_location, precond_scale, precond_prior_prec,
                obs_prec=obs_prec
            )
            return hessian_matvec(v)

        return precond_scale, compute_negative_logp, compute_negative_grad, precond_hessian_matvec

    def choose_optim_method_and_options(
            self, optim_maxiter, use_newton_method, require_trust_region, n_param):

        if optim_maxiter is None:
            if use_newton_method:
                optim_maxiter = 15
            else:
                optim_maxiter = 250

        optim_options = {'maxiter': optim_maxiter}
        tol = 10 ** -6 / np.sqrt(n_param)  # In analogy with the CG-sampler.

        if not use_newton_method:
            optim_method = 'L-BFGS-B'
            optim_options['gtol'] = tol
            optim_options['maxcor'] = 200
        else:
            if require_trust_region:
                optim_method = 'trust-ncg'
                # Start with a generous trust radius as Newton iterations without
                # constraints should be fine.
                init_trust_radius = 1.96 * np.sqrt(n_param)
                optim_options.update({
                    'gtol': tol,
                    'initial_trust_radius': init_trust_radius,
                    'max_trust_radius': 4. * init_trust_radius
                })
            else:
                optim_method = 'Newton-CG'
                optim_options['xtol'] = tol

        return optim_method, optim_options


class StabilityEstimateStabilizer():
    """ Detect and adjust unusually large stability limit estimates. """

    def __init__(self, n_warmup=100):
        self.stability_estimate = []
        self.n_update = 0
        self.n_warmup = n_warmup

    def update(self, estimate):
        self.stability_estimate.append(estimate)
        self.n_update += 1

    def stabilize(self, estimate):

        if self.n_update < self.n_warmup:
            return estimate

        gaussian_cdf_at_onestd = .8414
        stability_estimate = np.array(self.stability_estimate[:self.n_update])
        cdf_at_estimate = np.mean(stability_estimate < estimate)

        if cdf_at_estimate <= gaussian_cdf_at_onestd:
            stabilized_est = estimate
        else:
            stability_median = np.median(stability_estimate)
            stability_at_gaussian_cdf = np.quantile(
                stability_estimate, gaussian_cdf_at_onestd
            )
            one_std_dist = stability_at_gaussian_cdf - stability_median
            gaussian_dist_above_onestd = sp.stats.norm.ppf(cdf_at_estimate) - 1.
            gaussian_dist_above_onestd = min(2., gaussian_dist_above_onestd)
                # In case the empirical CDF of the estimate is too close to 1.
            stabilized_est = stability_at_gaussian_cdf \
                    + one_std_dist * gaussian_dist_above_onestd

        return stabilized_est



import numpy as np

def numerical_grad_is_close(
        f, x, atol=10E-6, rtol=10E-6, dx=10E-6,
        subset_index=None, return_grad=False):
    """
    Compare the computed gradient to a centered finite difference approximation.

    Params:
    -------
    f : callable
        Returns a value of a function and its gradient
    """
    if subset_index is None:
        subset_index = np.arange(len(x))

    x = np.array(x, ndmin=1)
    grad_est = np.zeros(len(subset_index))
    for i in range(len(subset_index)):
        x_minus = x.copy()
        x_minus[subset_index[i]] -= dx
        x_plus = x.copy()
        x_plus[subset_index[i]] += dx
        f_minus, _ = f(x_minus)
        f_plus, _ = f(x_plus)
        grad_est[i] = (f_plus - f_minus) / (2 * dx)

    _, grad = f(x)
    is_close = np.allclose(grad[subset_index], grad_est, atol=atol, rtol=rtol)

    if return_grad:
        return is_close, grad[subset_index], grad_est
    else:
        return is_close


def numerical_direc_deriv_is_close(
        f, x, hess_matvec, n_direction=10,
        atol=10E-6, rtol=10E-6, dx=10E-6, seed=None):
    """
    Compare analytically computed directional derivatives of the gradient of 'f'
    (i.e. the Hessian of 'f' applied to vectors) to its numerical approximations.

    Params:
    -------
    f : callable
        Returns a value of a function and its gradient
    """

    x = np.array(x, ndmin=1)

    np.random.seed(seed)
    all_matched = True

    for i in range(n_direction):

        v = np.random.randn(len(x))
        v /= np.sqrt(np.sum(v ** 2))
        _, grad_minus = f(x - dx * v)
        _, grad_plus = f(x + dx * v)
        direc_deriv_est = (grad_plus - grad_minus) / (2 * dx)
        direc_deriv = hess_matvec(v)

        if not np.allclose(direc_deriv, direc_deriv_est, atol=atol, rtol=rtol):
            all_matched = False
            break

    return all_matched



import numpy as np
import math
import time
from .stepsize_adapter import HamiltonianBasedStepsizeAdapter, initialize_stepsize
from .util import warn_message_only
from .dynamics import HamiltonianDynamics


dynamics = HamiltonianDynamics()
integrator = dynamics.integrate
compute_hamiltonian = dynamics.compute_hamiltonian
draw_momentum = dynamics.draw_momentum


def generate_samples(
        f, q0, n_burnin, n_sample, nstep_range, dt_range=None,
        seed=None, n_update=0, adapt_stepsize=False, target_accept_prob=.9,
        final_adaptsize=.05):
    """ Run HMC and return samples and some additional info. """

    if seed is not None:
        np.random.seed(seed)

    q = q0
    logp, grad = f(q)

    if np.isscalar(dt_range):
        dt_range = np.array(2 * [dt_range])

    elif dt_range is None:
        p = draw_momentum(len(q))
        logp_joint0 = - compute_hamiltonian(logp, p)
        dt = initialize_stepsize(
            lambda dt: compute_onestep_accept_prob(dt, f, q, p, grad, logp_joint0)
        )
        dt_range = dt * np.array([.8, 1.0])
        adapt_stepsize = True

    if np.isscalar(nstep_range):
        nstep_range = np.array(2 * [nstep_range])

    max_stepsize_adapter = HamiltonianBasedStepsizeAdapter(
        init_stepsize=1., target_accept_prob=target_accept_prob,
        reference_iteration=n_burnin, adaptsize_at_reference=final_adaptsize
    )

    if n_update > 0:
        n_per_update = math.ceil((n_burnin + n_sample) / n_update)
    else:
        n_per_update = float('inf')

    samples = np.zeros((len(q), n_sample + n_burnin))
    logp_samples = np.zeros(n_sample + n_burnin)
    accept_prob = np.zeros(n_sample + n_burnin)

    tic = time.time()  # Start clock
    use_averaged_stepsize = False
    for i in range(n_sample + n_burnin):
        dt = np.random.uniform(dt_range[0], dt_range[1])
        dt *= max_stepsize_adapter.get_current_stepsize(use_averaged_stepsize)
        nstep = np.random.randint(nstep_range[0], nstep_range[1] + 1)
        q, info = generate_next_state(
            f, dt, nstep, q, logp0=logp, grad0=grad
        )
        logp, grad, pathlen, accept_prob[i] = (
            info[key] for key in ['logp', 'grad', 'n_grad_evals', 'accept_prob']
        )
        if i < n_burnin and adapt_stepsize:
            max_stepsize_adapter.adapt_stepsize(info['hamiltonian_error'])
        elif i == n_burnin - 1:
            use_averaged_stepsize = True
        samples[:, i] = q
        logp_samples[i] = logp
        if (i + 1) % n_per_update == 0:
            print('{:d} iterations have been completed.'.format(i + 1))

    toc = time.time()
    time_elapsed = toc - tic

    return samples, logp_samples, accept_prob, time_elapsed


def compute_onestep_accept_prob(dt, f, q0, p0, grad0, logp_joint0):
    _, p, logp, _ = integrator(f, dt, q0, p0, grad0)
    logp_joint = - compute_hamiltonian(logp, p)
    accept_prob = np.exp(logp_joint - logp_joint0)
    return accept_prob


def generate_next_state(
        f, dt, n_step, q0,
        p0=None, logp0=None, grad0=None, hamiltonian_tol=100.):

    n_grad_evals = 0

    if (logp0 is None) or (grad0 is None):
        logp0, grad0 = f(q0)
        n_grad_evals += 1

    if p0 is None:
        p0 = draw_momentum(len(q0))

    log_joint0 = - compute_hamiltonian(logp0, p0)

    q, p, logp, grad, simulation_info = simulate_dynamics(
        f, dt, n_step, q0, p0, logp0, grad0, hamiltonian_tol
    )
    n_grad_evals += simulation_info['n_grad_evals']
    instability_detected = simulation_info['instability_detected']

    if instability_detected:
        acceptprob = 0.
        hamiltonian_error = - float('inf')
    else:
        log_joint = - compute_hamiltonian(logp, p)
        hamiltonian_error = log_joint - log_joint0
        acceptprob = min(1, np.exp(hamiltonian_error))

    accepted = acceptprob > np.random.rand()
    if not accepted:
        q = q0
        logp = logp0
        grad = grad0

    info = {
        'logp': logp,
        'grad': grad,
        'accepted': accepted,
        'accept_prob': acceptprob,
        'hamiltonian_error': hamiltonian_error,
        'instability_detected': instability_detected,
        'n_grad_evals': n_grad_evals
    }

    return q, info


def simulate_dynamics(f, dt, n_step, q0, p0, logp0, grad0, hamiltonian_tol=float('inf')):

    n_grad_evals = 0
    instability_detected = False

    # Keep track of Hamiltonians along the trajectory.
    hamiltonians = np.full(n_step + 1, float('nan'))
    hamiltonian = compute_hamiltonian(logp0, p0)
    hamiltonians[0] = hamiltonian
    min_h, max_h = 2 * [hamiltonian]

    q, p, logp, grad = q0, p0, logp0, grad0
    if n_step == 0:
        warn_message_only("The number of integration steps was set to be 0.")

    for i in range(n_step):
        q, p, logp, grad \
            = integrator(f, dt, q, p, grad)
        hamiltonian = compute_hamiltonian(logp, p)
        hamiltonians[i + 1] = hamiltonian
        min_h, max_h = update_running_minmax(min_h, max_h, hamiltonian)
        n_grad_evals += 1
        instability_detected \
            = math.isinf(logp) or (max_h - min_h) > hamiltonian_tol
        if instability_detected:
            warn_message_only(
                "Numerical integration became unstable while simulating the "
                "HMC trajectory."
            )
            break

    info = {
        'energy_trajectory': hamiltonians,
        'n_grad_evals': n_grad_evals,
        'instability_detected': instability_detected,
    }

    return q, p, logp, grad, info


def update_running_minmax(running_min, running_max, curr_val):
    running_min = min(running_min, curr_val)
    running_max = max(running_max, curr_val)
    return running_min, running_max



from distutils.core import setup, Extension
from Cython.Build import cythonize
import numpy as np

ext_modules = [
    Extension(
        "polya_gamma",
        ["polya_gamma.pyx"],
        include_dirs=[np.get_include()]
    )
]

setup(
    ext_modules = cythonize(ext_modules)
)



import numpy as np
import scipy as sp


def compute_credible_intervals(
        samples, credible_level=.95, use_median=False):
    tail_percentile = 100 * (1 - credible_level) / 2
    compute_middle = np.median if use_median else np.mean
    middle = compute_middle(samples, -1)
    upper = np.percentile(samples, 100 - tail_percentile, axis=-1)
    lower = np.percentile(samples, tail_percentile, axis=-1)
    return middle, lower, upper


def find_signif_coef(coef_samples, signif_level=.05, n_burnin=0):
    """
    Returns the indices of the cofficients whose one-sided tail probability
    is larger than 1 - signif_level / 2.
    """

    tail_prob = compute_tail_prob(coef_samples, n_burnin)
    top_coef_index, = np.nonzero(tail_prob > (1 - signif_level / 2))
    tail_prob = tail_prob[top_coef_index]

    # Sort the indices in terms of the tail probabilities.
    sorting_index = np.argsort(tail_prob)[::-1]
    top_coef_index = top_coef_index[sorting_index]
    tail_prob = tail_prob[sorting_index]

    # Break the ties in terms of the posterior mean.
    tie_indicator = (tail_prob == 1.)
    tied_indices = top_coef_index[tie_indicator]
    post_mean_magnitude = np.abs(
        np.mean(coef_samples[tied_indices, n_burnin:], -1))
    sorting_index = np.flip(np.argsort(post_mean_magnitude))
    top_coef_index[tie_indicator] = tied_indices[sorting_index]
    tail_prob[tie_indicator] = tail_prob[tie_indicator][sorting_index]

    return top_coef_index, tail_prob


def compute_tail_prob(coef_samples, n_burnin=0):
    tail_prob = np.max(np.vstack((
        np.mean(coef_samples[:, n_burnin:] > 0, 1),
        np.mean(coef_samples[:, n_burnin:] < 0, 1)
    )), 0)
    return tail_prob


# def compute_ave_treatment_effect(
#         treatment, X_interaction, treatment_samples, interaction_samples):
#     treatment_effect_samples = np.zeros(treatment_samples.size)
#     treatment_freq = np.mean(treatment)
#     for i in range(treatment_samples.size):
#         treatment_effect_samples[i] \
#             = treatment_freq * treatment_samples[i] \
#               + X_interaction.dot(interaction_samples[:, i]).mean()
#     return treatment_effect_samples


def compute_ave_treatment_effect(
        treatment_coef_samples, interaction_coef_samples, treatment, X_main):
    n_mcmc_sample = treatment_coef_samples.size
    treatment_effect_samples = np.squeeze(treatment_coef_samples)
    for i in range(n_mcmc_sample):
        treatment_effect_samples[i] \
            += X_main.dot(interaction_coef_samples[:, i]).mean()
    return treatment_effect_samples



from cpython.pycapsule cimport PyCapsule_GetPointer
from numpy.random cimport bitgen_t
from numpy.random.bit_generator cimport BitGenerator, bitgen_t
from numpy.random.c_distributions cimport random_standard_normal


cdef double random_normal(BitGenerator bit_generator):
    """
    Generate a random value from a standard normal distribution.

    Parameters
    ----------
    bit_generator : BitGenerator
        Numpy BitGenerator object. This object is *not* locked during generation since the
        sampling runs on a single thread and performance is much better without locking/releasing.

    Returns
    -------
    double
        Random number.
    """
    cdef bitgen_t *rng
    cdef const char *capsule_name = "BitGenerator"
    capsule = bit_generator.capsule
    rng = <bitgen_t *> PyCapsule_GetPointer(capsule, capsule_name)
    return random_standard_normal(rng)



import numpy as np
import scipy as sp
import scipy.sparse

def generate_gaussian_with_weight(X, obs_prec, prior_prec_sqrt, z, rand_gen=None):
    """
    Generate a multi-variate Gaussian with the mean mu and covariance Sigma of the form
        mu = Sigma z,
        Sigma^{-1} = X' diag(obs_prec) X + diag(prior_prec_sqrt) ** 2,

    Parameters
    ----------
        obs_prec : 1-d numpy array
        prior_prec_sqrt : 1-d numpy array
    """

    diag_sqrt = prior_prec_sqrt.copy()
    fisher_info = X.compute_fisher_info(weight=obs_prec, diag_only=True)
    fisher_info[fisher_info < 0.] = 0.
    fisher_info_sqrt = np.sqrt(fisher_info)
    has_pos_prior_prec = (prior_prec_sqrt > 0)
    has_zero_prior_prec = np.logical_not(has_pos_prior_prec)
    diag_sqrt[has_pos_prior_prec] *= np.sqrt(
        1. + (fisher_info_sqrt[has_pos_prior_prec] / diag_sqrt[has_pos_prior_prec]) ** 2
    )
    diag_sqrt[has_zero_prior_prec] = fisher_info_sqrt[has_zero_prior_prec]
    inv_sqrt_diag_scale = 1 / diag_sqrt
    Phi_scaled = inv_sqrt_diag_scale[:, np.newaxis] \
        * X.compute_fisher_info(obs_prec) * inv_sqrt_diag_scale[np.newaxis, :]
    Phi_scaled += np.diag((inv_sqrt_diag_scale * prior_prec_sqrt) ** 2)
    Phi_scaled_chol = sp.linalg.cholesky(Phi_scaled)
    mu = sp.linalg.cho_solve((Phi_scaled_chol, False), inv_sqrt_diag_scale * z)
    if rand_gen is None:
        gaussian_vec = np.random.randn(X.shape[1])
    else:
        gaussian_vec = rand_gen.np_random.randn(X.shape[1])
    beta_scaled = mu + sp.linalg.solve_triangular(
        Phi_scaled_chol, gaussian_vec, lower=False
    )
    beta = inv_sqrt_diag_scale * beta_scaled

    return beta



import numpy as np
from .polya_gamma import PolyaGammaDist
from .tilted_stable import ExpTiltedStableDist

class BasicRandom():
    """
    Generators of random variables from the basic distributions used in
    Bayesian sparse regression.
    """

    def __init__(self, seed=None):
        self.np_random = np.random
        self.pg = PolyaGammaDist()
        self.ts = ExpTiltedStableDist()
        self.set_seed(seed)

    def set_seed(self, seed):
        self.np_random.seed(seed)
        pg_seed = np.random.randint(1, 1 + np.iinfo(np.int32).max)
        ts_seed = np.random.randint(1, 1 + np.iinfo(np.int32).max)
        self.pg.set_seed(pg_seed)
        self.ts.set_seed(ts_seed)

    def get_state(self):
        rand_gen_state = {
            'numpy' : self.np_random.get_state(),
            'tilted_stable' : self.ts.get_state(),
            'polya_gamma' : self.pg.get_state()
        }
        return rand_gen_state

    def set_state(self, rand_gen_state):
        self.np_random.set_state(rand_gen_state['numpy'])
        self.ts.set_state(rand_gen_state['tilted_stable'])
        self.pg.set_state(rand_gen_state['polya_gamma'])

    def polya_gamma(self, shape, tilt):
        return self.pg.rand_polyagamma(shape, tilt)

    def tilted_stable(self, char_exponent, tilt):
        return self.ts.sample(char_exponent, tilt)



from .abstract_model import AbstractModel
import numpy as np
import scipy as sp
from warnings import warn


class CoxModel(AbstractModel):

    def __init__(self, event_time, censoring_time, design):
        """

        Parameters
        ----------
        event_time : numpy array
            The lowest values indicate the earliest events. float('inf')
            indicates right-censoring.
        censoring_time : numpy array
            float('inf') indicates uncensored observations.
        """

        if np.any(event_time[:-1] > event_time[1:]):
            raise ValueError(
                "The observations need to be sorted so that the event times are "
                "in the increasing order, from the earliest to last events."
            )

        if np.any(censoring_time[:-1] < censoring_time[1:]):
            raise ValueError(
                "The observations need to be sorted so that the censoring times "
                "are in the decreasing order, from uncensored, last censored, "
                "to the earliest censored."
            )

        n_event = len(event_time) - np.sum(np.isinf(event_time))
        risk_set_start_index, risk_set_end_index = \
            self._find_risk_set_index(
                event_time[:n_event],
                np.flip(censoring_time[n_event:])
            )
        n_appearance = CoxModel.count_risk_set_appearance(
            len(event_time), risk_set_start_index, risk_set_end_index
        )
        if not np.all(n_appearance >= 1):
            raise ValueError(
                "Some individuals never appear in the risk set. They have to be"
                "removed before using the CoxModel class.")

        self.n_event = n_event
        self.event_time = event_time
        self.censoring_time = censoring_time
        self.n_appearance_in_risk_set = n_appearance
        self.risk_set_start_index = risk_set_start_index
        self.risk_set_end_index = risk_set_end_index
        self.design = design
        self.name = 'cox'

    @staticmethod
    def preprocess_data(event_time, censoring_time, X):
        event_time, censoring_time, X = \
            CoxModel._permute_observations_by_event_and_censoring_time(
                event_time, censoring_time, X
            )
        event_time, censoring_time, X = \
            CoxModel._drop_uninformative_observations(
                event_time, censoring_time, X
            )
        return event_time, censoring_time, X

    @staticmethod
    def _permute_observations_by_event_and_censoring_time(
            event_time, censoring_time, X):
        """
        Permute the observations so that they are ordered from the earliest
        to the last to experience events, followed by the last censored to the
        earliest censored.

        Params
        ------
        event_time : numpy array
            The lowest values indicate the earliest events. float('inf')
            indicates right-censoring.
        X : numpy array or scipy sparse matrix
        """
        if not np.all(np.equal(
                event_time == float("inf"),
                censoring_time < float('inf')
            )):
            raise ValueError(
                "Either event or censoring time must be infinity for each "
                "observation."
            )

        is_sorted = (
            np.all(event_time[:-1] <= event_time[1:])
            and np.all(censoring_time[:-1] >= censoring_time[1:])
        )
        if is_sorted:
            return event_time, censoring_time, X

        warn(
            "The observations and design matrix will be sorted so that the event "
            "times are in the ascending order and censoring times in the descending order."
        )

        n_event = np.sum(event_time < float('inf'))
        event_rank = CoxModel.np_rank_by_value(event_time)
        censoring_rank = CoxModel.np_rank_by_value(censoring_time)
        sort_ind = np.concatenate((
            np.argsort(event_rank)[:n_event],
            np.argsort(- censoring_rank)[n_event:]
        ))
        assert len(np.unique(sort_ind)) == len(sort_ind)

        event_time = event_time[sort_ind]
        censoring_time = censoring_time[sort_ind]
        if sp.sparse.issparse(X):
            X = X.tocsr()[sort_ind, :]
        else:
            X = X[sort_ind, :]

        return event_time, censoring_time, X

    @staticmethod
    def _drop_uninformative_observations(event_time, censoring_time, X):

        finite_event_time = event_time[event_time < float('inf')]
        finite_censoring_time = censoring_time[censoring_time < float('inf')]

        # Exclude those censored before the first event.
        is_uninformative = (censoring_time < np.min(event_time))

        if np.any(is_uninformative):
            warn(
                "Some observations do not contribute to the likelihood, so "
                "they are being removed."
            )
            is_informative = np.logical_not(is_uninformative)
            event_time = event_time[is_informative]
            censoring_time = censoring_time[is_informative]
            X = X[is_informative, :]

        return event_time, censoring_time, X

    @staticmethod
    def np_rank_by_value(arr):
        sort_arguments = np.argsort(arr)
        rank = np.arange(len(arr))[np.argsort(sort_arguments)]
        return rank.astype('float')

    @staticmethod
    def count_risk_set_appearance(n_obs, start_index, end_index):
        """ This function assumes that the observations are already sorted in
        the way required by the class. """

        # The calculation can be done more efficiently.
        n_appearance = np.zeros(n_obs, dtype=np.int)
        for i in range(len(start_index)):
            if start_index[i] <= end_index[i]:
                n_appearance[start_index[i]:(end_index[i] + 1)] += 1
        return n_appearance

    def _find_risk_set_index(self, event_time, censoring_time):
        """ The parameters are assumed to have 'inf' removed and in the ascending order. """

        n_event = len(event_time)
        start_index = np.zeros(n_event, dtype=np.int)
        for i in range(1, n_event):
            if event_time[i - 1] == event_time[i]:
                start_index[i] = start_index[i - 1]
            else:
                start_index[i] = i

        n_censored = np.array([
            np.searchsorted(censoring_time, t) for t in event_time
        ], dtype=np.int) # Tied censoring time is considered to be in the risk set.
        end_index = len(event_time) + len(censoring_time) - 1 - n_censored

        return start_index, end_index

    def compute_loglik_and_gradient(self, beta, loglik_only=False):

        grad = None # defalt return value

        log_rel_hazard, rel_hazard, hazard_sum_over_risk_set \
            = self._compute_relative_hazard(beta)
        if np.any(hazard_sum_over_risk_set == 0.):
            loglik = - float('inf')
            return loglik, grad

        loglik = np.sum(
            log_rel_hazard[:self.n_event] - np.log(hazard_sum_over_risk_set)
        )

        if not loglik_only:
            hazard_matrix = self._HazardMultinomialProbMatrix(
                rel_hazard, hazard_sum_over_risk_set,
                self.risk_set_start_index, self.risk_set_end_index, self.n_appearance_in_risk_set
            )
            v = np.zeros(self.design.shape[0])
            v[:self.n_event] = 1
            v -= hazard_matrix.sum_over_events()
            grad = self.design.Tdot(v)

        return loglik, grad

    def _compute_relative_hazard(self, beta):

        log_rel_hazard = self.design.dot(beta)
        log_rel_hazard = CoxModel._shift_log_hazard(log_rel_hazard)

        rel_hazard = np.exp(log_rel_hazard)

        hazard_sum_over_risk_set = self._sum_over_start_end(
            rel_hazard, self.risk_set_start_index, self.risk_set_end_index
        )

        return log_rel_hazard, rel_hazard, hazard_sum_over_risk_set

    @staticmethod
    def _sum_over_start_end(arr, start_index, end_index):
        """
        Returns
        -------
        numpy array whose k-th element equals
            np.sum(arr[start_index[k]:(1 + end_index[k])])
        """
        sum_from_right = \
            np.cumsum(arr[start_index[-1]:])[end_index - start_index[-1]]
        sum_from_left = np.concatenate((
            CoxModel.np_reverse_cumsum(arr[:start_index[-1]]), [0]
        ))
        total_sum = sum_from_right + sum_from_left
        return total_sum

    @staticmethod
    def np_reverse_cumsum(arr):
        return np.cumsum(arr[::-1])[::-1]

    @staticmethod
    def _shift_log_hazard(log_hazard_rate, log_offset=0):
        """
        Shift the values so that the max value equals 'log_offset' to
        prevent numerical under / over-flow before taking exponential.
        """
        log_hazard_rate += log_offset - np.max(log_hazard_rate)
        return log_hazard_rate

    def compute_hessian(self, beta):
        raise NotImplementedError()

    def get_hessian_matvec_operator(self, beta):

        _, rel_hazard, hazard_sum_over_risk_set \
            = self._compute_relative_hazard(beta)
        if np.any(hazard_sum_over_risk_set == 0.):
            raise ValueError(
                'Hessian operator cannot be computed likely due to an '
                'unreasonable value of regression coefficients. This could '
                'be caused by the likelihood and prior both being too weak '
                'or by a poor initialization of the Markov chain.'
            )
        W = self._HazardMultinomialProbMatrix(
            rel_hazard, hazard_sum_over_risk_set,
            self.risk_set_start_index, self.risk_set_end_index, self.n_appearance_in_risk_set
        )
        def hessian_op(beta):
            X_beta = self.design.dot(beta)
            result_vec = - self.design.Tdot(
                W.sum_over_events() * X_beta - W.Tdot(W.dot(X_beta))
            )
            return result_vec

        return hessian_op

    @staticmethod
    def simulate_outcome(X, beta, censoring_frac=.9, seed=None):
        """
        Simulate an outcome from a constant baseline hazard model i.e. the
        survival time is exponential.
        """
        if seed is not None:
            np.random.seed(seed)

        log_hazard_rate = X.dot(beta)
        log_hazard_rate = CoxModel._shift_log_hazard(log_hazard_rate)
        hazard_rate = np.exp(log_hazard_rate)
        event_time = np.random.exponential(scale=hazard_rate ** -1)

        scale = CoxModel._solve_for_exp_scale(
            np.quantile(event_time, 1 - censoring_frac), 1 - censoring_frac
        )
        censoring_time = np.random.exponential(
            scale=scale * np.ones(len(hazard_rate))
        )
        censoring_time[event_time < censoring_time] = float("inf")
        event_time[event_time >= censoring_time] = float('inf')

        return event_time, censoring_time

    @staticmethod
    def _solve_for_exp_scale(t, prob):
        """
        Computes the scale of an exponential random variable Z such that
            P(Z < t) == prob
        """
        return - t / np.log(1 - prob)

    class _HazardMultinomialProbMatrix():
        """
        Defines operations by a matrix whose each row represents the conditional
        probabilities of the event happening to the individuals in the risk set.
        """

        def __init__(self, rel_hazard, hazard_sum_over_risk_set,
                     risk_set_start_index, risk_set_end_index, n_appearance_in_risk_set):
            self.rel_hazard = rel_hazard
            self.hazard_sum_over_risk_set = hazard_sum_over_risk_set
            self.risk_set_start_index = risk_set_start_index
            self.risk_set_end_index = risk_set_end_index
            self.n_appearance_in_risk_set = n_appearance_in_risk_set
            self.n_event = len(hazard_sum_over_risk_set)


        def sum_over_events(self):
            """
            Returns the same value as the row sum of the explicitly computed
            the matrix (e.g. via the 'compute_matrix' method) but do it more
            efficiently.
            """
            normalizer_cumsum = np.cumsum(self.hazard_sum_over_risk_set ** -1)
            row_sum = normalizer_cumsum[self.n_appearance_in_risk_set - 1] \
                      * self.rel_hazard
            return row_sum

        def dot(self, v):
            return self.hazard_sum_over_risk_set ** - 1 * CoxModel._sum_over_start_end(
                self.rel_hazard * v, self.risk_set_start_index, self.risk_set_end_index
            )

        def Tdot(self, v):
            partial_inner_prod = np.cumsum(self.hazard_sum_over_risk_set ** -1 * v)
            return self.rel_hazard \
                   * partial_inner_prod[self.n_appearance_in_risk_set - 1]

        def compute_matrix(self):
            multinomial_prob = np.outer(
                self.hazard_sum_over_risk_set ** -1,
                self.rel_hazard
            )
            multinomial_prob = np.triu(multinomial_prob)
            for i in range(1, multinomial_prob.shape[0] + 1):
                if (self.risk_set_end_index[-i] + 1) >= len(self.rel_hazard):
                    break
                multinomial_prob[-i, (self.risk_set_end_index[-i] + 1):] = 0

            return multinomial_prob


import numpy as np
from .abstract_matrix import AbstractDesignMatrix


class DenseDesignMatrix(AbstractDesignMatrix):
    
    def __init__(self, X, center_predictor=False, add_intercept=True,
                 copy_array=False):
        """
        Params:
        ------
        X : numpy array
        """
        if copy_array:
            X = X.copy()
        super().__init__()
        X = self.remove_intercept_indicator(X)
        if center_predictor:
            X -= np.mean(X, axis=0)[np.newaxis, :]
        if add_intercept:
            X = np.hstack((np.ones((X.shape[0], 1)), X))
        self.X = X
        self.intercept_added = add_intercept
        self.centered = center_predictor

    @property
    def shape(self):
        return self.X.shape

    @property
    def is_sparse(self):
        return False

    def dot(self, v):

        if self.memoized and np.all(self.v_prev == v):
            return self.X_dot_v

        result = self.X.dot(v)
        if self.memoized:
            self.X_dot_v = result
            self.v_prev = v
        self.dot_count += 1

        return result

    def Tdot(self, v):
        self.Tdot_count += 1
        return self.X.T.dot(v)

    def compute_fisher_info(self, weight, diag_only=False):
        if diag_only:
            return np.sum(weight[:, np.newaxis] * self.X ** 2, 0)
        else:
            return self.X.T.dot(weight[:, np.newaxis] * self.X)

    def toarray(self):
        return self.X

    def extract_matrix(self, order=None):
        return self.X



import matplotlib.pyplot as plt
import numpy as np


def init_figure(figsize=(8, 5), fontsize=18):
    plt.figure(figsize=figsize)
    plt.rcParams['font.size'] = fontsize


def savefig(filename, bbox_inches='tight', pad_inches=0, **kwargs):
    plt.savefig(
        filename, bbox_inches=bbox_inches, pad_inches=pad_inches, **kwargs
    )


def add_diag_refline(
        x, y, linestyle='--', equal_axis=True, zorder=0, label='x = y',
        **kwargs):
    min_max = [
        min(np.min(x), np.min(y)),
        max(np.max(x), np.max(y))
    ]
    plt.plot(
        min_max, min_max,
        linestyle=linestyle, zorder=zorder, label=label, **kwargs
    )
    if equal_axis:
        plt.gca().set_aspect('equal', 'box')


def add_line_at_zero(horizontal=True, vertical=False, labels=[None, None],
                     color='grey', linestyle='--', linewidth=.8):
    """
    Parameters
    ----------
    labels: list of str e.g. ['y = 0', 'x = 0']
    """
    if horizontal:
        plt.axhline(
            y=0, color=color, linestyle=linestyle, linewidth=linewidth, label=labels[0]
        )
    if vertical:
        plt.axvline(
            x=0, color=color, linestyle=linestyle, linewidth=linewidth, label=labels[1]
        )
        
        
def remove_figure_box_edges(sides=['left', 'right', 'top', 'bottom']):
    for side in sides:
        plt.gca().spines[side].set_visible(False)


def delineated_hist(x, bins, color):
    plt.hist(
        x, bins=bins, histtype='step', ec=color, lw=1.25
    )
    plt.hist(
        x, bins=bins, fc=color, alpha=.1
    )


import numpy as np
cimport numpy as np
import cython
cimport cython
from cython.parallel cimport prange

ctypedef np.int32_t INT_t
ctypedef np.float_t FLOAT_t
FLOAT = np.float64

def binary_matmul(X_csr, v):
  return c_binary_matmul_parallel(X_csr.indices, X_csr.indptr, v)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef c_binary_matmul(np.ndarray[INT_t, ndim=1] indices, np.ndarray[INT_t, ndim=1] indptr, np.ndarray[FLOAT_t, ndim=1] v):
    cdef int i, k
    cdef int m = indptr.shape[0] - 1
    cdef FLOAT_t val
    cdef np.ndarray[FLOAT_t, ndim=1] Xv = np.zeros(m, dtype=FLOAT)
    for i in range(m):
        val = 0
        for k in range(indptr[i], indptr[i + 1]):
            val += v[indices[k]]
        Xv[i] = val
    return Xv

@cython.boundscheck(False)
@cython.wraparound(False)
cdef c_binary_matmul_parallel(np.ndarray[INT_t, ndim=1] indices, np.ndarray[INT_t, ndim=1] indptr, np.ndarray[FLOAT_t, ndim=1] v):
    cdef int i, k
    cdef int m = indptr.shape[0] - 1
    cdef FLOAT_t val
    cdef np.ndarray[FLOAT_t, ndim=1] Xv = np.zeros(m, dtype=FLOAT)
    for i in prange(m, nogil=True):
        for k in range(indptr[i], indptr[i + 1]):
            Xv[i] += v[indices[k]]
    return Xv



import numpy as np
import math


"""
Defines a (numerical) Hamiltonian dynamics based on a Gaussian momentum and the 
velocity Verlet integrator. The code is written so that other integrators & 
momentum distributions can also be employed straightwardly.
"""

class HamiltonianDynamics():

    def __init__(self, mass=None):
        """
        Parameters
        ----------
        mass: None, numpy 1d array, or callable `mass(p, power)`
            If callable, should return a vector obtained by multiplying the
            vector p with matrix M ** power for power == -1 or power == 1/2.
            The matrix L corresponding to M ** 1/2 only needs to satisfy L L' = M.
            Passing M = None defaults to a dynamics with the identity mass matrix.
        """

        if mass is None:
            mass_operator = lambda p, power: p
        elif isinstance(mass, np.ndarray):
            sqrt_mass = np.sqrt(mass)
            inv_mass = 1 / mass
            def mass_operator(p, power):
                if power == -1:
                    return inv_mass * p
                elif power == 1 / 2:
                    return sqrt_mass * p
        elif callable(mass):
            mass_operator = mass
        else:
            raise ValueError("Unsupported type for the mass matrix.")

        self.integrator = velocity_verlet
        self.momentum = GaussianMomentum(mass_operator)

    def integrate(self, f, dt, q, p, grad):
        q, p, logp, grad \
            = self.integrator(f, self.momentum.get_grad, dt, q, p, grad)
        return q, p, logp, grad

    def draw_momentum(self, n_param):
        return self.momentum.draw_random(n_param)

    def compute_hamiltonian(self, logp, p):
        potential = - logp
        kinetic = - self.momentum.get_logp(p)
        return potential + kinetic

    def convert_to_velocity(self, p):
        return - self.momentum.get_grad(p)


def velocity_verlet(
        get_position_logp_and_grad, get_momentum_grad, dt, q, p, position_grad
    ):
    p = p + 0.5 * dt * position_grad
    q = q - dt * get_momentum_grad(p)
    position_logp, position_grad = get_position_logp_and_grad(q)
    if math.isfinite(position_logp):
        p += 0.5 * dt * position_grad
    return q, p, position_logp, position_grad


class GaussianMomentum():

    def __init__(self, mass=None):
        self.mass = mass

    def draw_random(self, n_param):
        p = self.mass(np.random.randn(n_param), 1/2)
        return p

    def get_grad(self, p):
        return - self.mass(p, -1)

    def get_logp(self, p):
        return - 0.5 * np.dot(p, self.mass(p, -1))


import cProfile, sys, time
import random
import numpy as np
from polya_gamma import rand_polya_gamma


def generate_samples(tilt):
    n_repetition = 10 ** 6
    for i in range(n_repetition):
        rand_polya_gamma(tilt)

if __name__ == '__main__':
    random.seed(0)
    for tilt in 10. ** np.array([-4, 0, 4]):
        print("Generating PG(1, {:.2g})...\n".format(tilt))
        if len(sys.argv) > 1 and sys.argv[1] == 'profile':
            cProfile.runctx("generate_samples(tilts)", globals(), locals(), sort='time')
            print("\n")
        else:
            start = time.time()
            generate_samples(tilt)
            elapsed = time.time() - start
            print("Took {:.3g} seconds.\n".format(elapsed))


"""

References:
----------
Philip Heidelberger and Peter D. Welch (1983). Simulation run length control
in the presence of an initial transient. Operations Research.

Sandor Csorgo and Julian J. Faraway (1996). The exact and asymptotic distributions
of Cramér-von Mises statistics. Journal of the Royal Statistical Society: Series B.

"""

import numpy as np
from math import pi, ceil, sqrt, exp, gamma, factorial
from scipy.special import kv as modified_bessel_2
from . import estimate_ess


def is_stationarity(x, signif_level=.05):
    return calculate_p_value(x) < signif_level


def calculate_p_value(x, ess_chain_frac=.5):
    """ Calculate the p-value under stationarity using Cramer-von-Mises statistics.

    Parameters
    ----------
    x: 1d numpy array
    ess_chain_frac : float between 0 and 1
        Fraction of the tail of the chain used in estimating the effective
        sample size, which is needed in computing the test statistics.
    """
    if (type(x) is not np.ndarray) or x.ndim != 1:
        raise TypeError("The input must be a 1d numpy array.")
    cvm_stat = cramer_von_mises_statistic(x, ess_chain_frac)
    p_val = 1 - cramer_von_mises_cdf(cvm_stat)
    return p_val


def cramer_von_mises_statistic(x, ess_chain_frac=.5):
    bb_stat = _brownian_bridge_statistic(x, 1 - ess_chain_frac)
    cvm_stat = np.trapz(bb_stat ** 2, np.linspace(0, 1, len(bb_stat)))
    return cvm_stat


def _brownian_bridge_statistic(x, frac_discard=.5):
    """
    Parameters
    ----------
    frac_discard : float in [0, 1]
        The non-stationarity can inflate the estimate of the spectrum at zero.
        To avoid this, as proposed in Heidelberger and Welch (1983), discard
        the initial fraction of the time series `x` (and hope that the rest of
        the sequence looks more stationary).
    """
    n_discard = ceil(frac_discard * len(x))
    x_subseq = x[n_discard:]
    spectrum_at_zero = np.var(x_subseq) / estimate_ess(x_subseq, normed=True)
    cumsum = np.concatenate(([0], np.cumsum(x)))
    linear_interp = np.arange(len(x) + 1) * np.mean(x)
    stat = (cumsum - linear_interp) / np.sqrt(len(x) * spectrum_at_zero)
    return stat


def cramer_von_mises_cdf(x, n_summand=None):
    """
    Computes the cumulative distribution function of the (asymptotic)
    Cramer-von-Mises statistics (the integral of a squared Brownian bridge
    process) via a series expansion formula.

    Parameters
    ----------
    x : scalar
    n_summand : optional, int
        The R coda (as of ver 0.19-1) uses 4, but we need more terms for a large
        value of `x`.
    """
    if n_summand is None:
        n_summand = 5 if x < 10 else 10
    cum_density = 0
    for k in range(n_summand):
        cum_density += _cvm_cdf_summand(x, k)
    return cum_density


def _cvm_cdf_summand(x, k):
    temp = (4 * k + 1) ** 2 / 16 / x
    kth_summand = (
        1 / pi ** (3 / 2) / sqrt(x)
        * gamma(k + .5) / factorial(k) * sqrt(4 * k + 1)
        * exp(-temp) * modified_bessel_2(.25, temp)
    ) # Equation (1.3) in Csorgo and Faraway (1996).
    return kth_summand



import numpy as np
import scipy as sp
import scipy.sparse
from bayesbridge.model import CoxModel


def simulate_outcome(X, beta, model, intercept=0., n_trial=None, seed=None):

    if seed is not None:
        np.random.seed(seed)

    if model == 'linear':
        sigma = 1.
        outcome = intercept + X.dot(beta) + sigma * np.random.randn(X.shape[0])
    elif model == 'logit':
        if n_trial is None:
            n_trial = np.ones(X.shape[0])
        prob = 1 / (1 + np.exp(- intercept - X.dot(beta)))
        n_success = np.random.binomial(n_trial.astype(np.int64), prob)
        outcome = (n_success, n_trial)
    elif model == 'cox':
        outcome = CoxModel.simulate_outcome(X, beta, censoring_frac=.5)
    else:
        raise NotImplementedError()

    return outcome

def simulate_design(
        n_obs, n_pred, binary_frac=0., categorical_frac=0.,
        corr_dense_design=False, binary_pred_freq=.1, n_category=5,
        shuffle_columns=False, seed=None, format_='sparse'
    ):

    if seed is not None:
        np.random.seed(seed)

    n_dense_pred = int(n_pred * (1 - binary_frac - categorical_frac))
    n_categorical_pred = int((n_pred * categorical_frac) / (n_category - 1))
    n_binary_pred = n_pred - n_dense_pred - n_categorical_pred * (n_category - 1)

    X_dense = simulate_dense_design(n_obs, n_dense_pred, corr_dense_design)

    if n_binary_pred + n_categorical_pred == 0:
        X = X_dense
    else:
        X_binary = simulate_binary_design(n_obs, n_binary_pred, binary_pred_freq)
        X_categorical = simulate_categorical_design(
            n_obs, n_categorical_pred, n_category
        )
        X = sp.sparse.hstack((
            sp.sparse.csr_matrix(X_dense), X_binary, X_categorical
        )).tocsr()

    if shuffle_columns:
        X = X[:, np.random.permutation(n_pred)]

    if format_ == 'sparse':
        X = sp.sparse.csr_matrix(X)
    elif sp.sparse.issparse(X):
        X = X.toarray()

    return X

def simulate_dense_design(n_obs, n_pred, corr_design, standardize=False):
    if corr_design:
        X = generate_corr_design(n_obs, n_pred)
    else:
        X = np.random.randn(n_obs, n_pred)
    if standardize:
        X = np_standardize(X)
    return X

def np_standardize(X, divide_by='std'):
    X = X - np.mean(X, axis=0)[np.newaxis, :]
    if divide_by == 'max':
        X = X / np.max(X, axis=0)[np.newaxis, :]
    else:
        X = X / np.std(X, axis=0)[np.newaxis, :]
    return X

def generate_corr_design(n_obs, n_pred, n_factor=None, max_sd=100, min_sd=1):
    """
    Each column is drawn from a Gaussian with a covariance proportional to
        I + F L F'
    where F is an orthogonal matrix of size p by n_factor and L is diagonal.
    """
    if n_factor is None:
        n_factor = min(100, int(n_pred / 2))
    factor, _ = np.linalg.qr(np.random.randn(n_pred, n_factor))
    principal_comp_sd = np.linspace(max_sd, min_sd, n_factor + 1)
    loading = principal_comp_sd[:n_factor] - min_sd
    X = np.dot(
        factor,
        loading[:, np.newaxis] * np.random.randn(n_factor, n_obs)
    ).T
    X += min_sd * np.random.randn(n_obs, n_pred)
    return X

def simulate_binary_design(n_obs, n_binary_pred, sparsity, max_freq_per_col=.5):
    """
    Returns a binary matrix where the non-zero frequency (on average) equals
    the value of 'sparsity'. Also, the non-zero frequency along each column is
    bounded by 'max_freq_per_col'.
    """
    if n_binary_pred == 0:
        return None

    a = .5
    b = a * (max_freq_per_col / sparsity - 1)
        # Solve a / (a + b) = sparsity / max_freq_per_col for 'b'.
    binary_freq = max_freq_per_col * np.random.beta(a, b, n_binary_pred)
    X = np.random.binomial(1, binary_freq, (n_obs, n_binary_pred))
    return X

def simulate_categorical_design(n_obs, n_categorical_pred, n_category=5):
    if n_categorical_pred == 0:
        return None

    X = sp.sparse.hstack([
        sp.sparse.csr_matrix(draw_categorical_pred(n_obs, n_category))
        for dummy in range(n_categorical_pred)
    ])
    return X

def draw_categorical_pred(n_obs, n_category):
    # Returns a matrix of size n by (n_category - 1).
    category_freq = np.random.dirichlet(
        np.ones(n_category - 1) / n_category
    )
    return np.random.multinomial(1, category_freq, n_obs)



import abc


class AbstractModel():

    @property
    def n_obs(self):
        return self.design.shape[0]

    @property
    def n_pred(self):
        return self.design.shape[1]

    @property
    def intercept_added(self):
        return self.design.intercept_added

    @abc.abstractmethod
    def compute_loglik_and_gradient(self, beta, loglik_only=False):
        pass

    @abc.abstractmethod
    def compute_hessian(self, beta):
        pass

    @abc.abstractmethod
    def get_hessian_matvec_operator(self, beta):
        pass

    @abc.abstractmethod
    def get_hessian_matvec_operator(self, beta):
        pass

    @abc.abstractmethod
    def calc_intercept_mle(self):
        """ Calculate MLE for intercept assuming other coefficients are zero. """
        pass

    @staticmethod
    @abc.abstractmethod
    def simulate_outcome():
        pass


import warnings


def simplified_format(
        message, category, filename, lineno, line=None):
    to_print = '{:s}:{:d}: {:s}: {:s}\n'.format(
        filename, lineno, category.__name__, str(message)
    )
    return to_print

warnings.formatwarning = simplified_format

def warn_message_only(message, category=UserWarning):
    warnings.warn(message, category, stacklevel=2)



import numpy as np
from Cython.Build import cythonize
from distutils.command.build_ext import build_ext
from numpy.distutils.misc_util import get_info
from os.path import dirname, join, abspath
from setuptools import setup, find_packages
from setuptools.extension import Extension

path = dirname(__file__)
src_dir = join(dirname(path), '..', 'src')
defs = [('NPY_NO_DEPRECATED_API', 0)]
inc_path = np.get_include()
lib_path = [abspath(join(np.get_include(), '..', '..', 'random', 'lib'))]
lib_path += get_info('npymath')['library_dirs']
np_libs = ['npyrandom', 'npymath']

class CustomBuildExtCommand(build_ext):
    """ build_ext command when numpy headers are needed. """
    def run(self):
        # Import numpy here, only when headers are needed
        import numpy as np
        self.include_dirs.append(np.get_include())
        build_ext.run(self)


ext_modules = [
    Extension(
        'bayesbridge.random.tilted_stable.tilted_stable',
        sources=['bayesbridge/random/tilted_stable/tilted_stable.pyx'],
    ),
    Extension(
        'bayesbridge.random.polya_gamma.polya_gamma',
        sources=['bayesbridge/random/polya_gamma/polya_gamma.pyx'],
    ),
    Extension(
        'bayesbridge.random.normal.normal',
        sources=['bayesbridge/random/normal/normal.pyx'],
        library_dirs=lib_path,
        libraries=np_libs,
        define_macros=defs,
    ),
    Extension(
        'bayesbridge.random.uniform.uniform',
        sources=['bayesbridge/random/uniform/uniform.pyx'],
        library_dirs=lib_path,
        libraries=np_libs,
        define_macros=defs,
    )
]

setup(
    name='bayesbridge',
    version='0.2.4',
    description=\
        'Generates posterior samples under Bayesian sparse regression based on '
        + 'the bridge prior using the CG-accelerated Gibbs sampler of Nishimura '
        + 'et. al. (2018). The linear and logistic model are currently supported.',
    url='https://github.com/aki-nishimura/bayes-bridge',
    author='Akihiko (Aki) Nishimura',
    author_email='aki.nishimura@jhu.edu',
    license='MIT',
    packages=find_packages(exclude=['tests', 'tests.*']),
    cmdclass={'build_ext': CustomBuildExtCommand},
    ext_modules=cythonize(ext_modules),
    setup_requires=['numpy>=1.19'],
    install_requires=[
        'numpy>=1.19', 'scipy'
    ],
    zip_safe=False
)



import numpy as np
import time


def random_walk_metropolis_sampler(
        f, x0, proposal_sd, n_samples, Sigma=None, seed=None, thin=1):
    """
    Generates samples via random-walk Metropolis.

    Params:
    ------
    f : function
        Computes the log density of the target density
    n_samples: int
        The number of samples to return. The total number of iteration is
        (thin * n_samples).
    """

    if seed is not None:
        np.random.seed(seed)

    if Sigma is not None:
        L = np.linalg.cholesky(Sigma)
    else:
        L = None

    n_param = len(np.atleast_1d(x0))
    mcmc_samples = np.zeros((n_param, n_samples))
    logp_samples = np.zeros(n_samples)
    accepted = np.zeros(thin)
    logp = f(x0)
    theta = x0
    mcmc_samples[:, 0] = theta
    logp_samples[0] = logp
    accept_rate = 0

    start_time = time.time()
    for mcmc_iter in range(1, n_samples):
        for j in range(thin):
            theta, logp, accepted[j], _ = random_walk_metropolis_step(
                f, theta, logp, proposal_sd, L, chol=True
            )
        mcmc_samples[:, mcmc_iter] = theta
        logp_samples[mcmc_iter] = logp
        weight = (mcmc_iter - 1) / mcmc_iter
        accept_rate = weight * accept_rate + (1 - weight) * np.mean(accepted)
    end_time = time.time()
    time_elapsed = end_time - start_time

    return mcmc_samples, accept_rate, logp_samples, time_elapsed


def random_walk_metropolis_step(f, x0, logp0, proposal_sd, Rho=None, chol=False):
    """
    Params
    ------
    f : function
        Computes the log density of the target density
    proposal_sd : scalar or vector
    Rho : matrix
        (Cholesky factor of) correlation matrix. If given, the proposal
        variance will be 'diag(proposal_sd) * Rho * diag(proposal_sd).'
    chol : bool
        If True, L = Rho is assumed to be the lower triangular cholesky
        factorization of the proposal covariance (which is given by L * L').
    """

    dx = np.random.randn(len(x0))
    if Rho is not None:
        L = Rho if chol else np.linalg.cholesky(Rho)
        dx = np.dot(L, dx)
    dx *= proposal_sd
    x = x0 + dx

    logp = f(x)
    accept_prob = min(1, np.exp(logp - logp0))
    accepted = accept_prob > np.random.uniform()
    if not accepted:
        x = x0
        logp = logp0

    return x, logp, accepted, accept_prob


class ProductGaussian:

    def __init__(self, marginal_sd):
        self.dim = len(np.atleast_1d(marginal_sd))
        self.sigma = marginal_sd

    def compute_logp(self, x):
        """ Computes the log-density (up to an additive constant). """
        return - 1 / 2 * np.sum((x / self.sigma) ** 2)

